/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.router;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.InputStream;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.common.binaryfile.GrpcFileUploadUtils.FileSender;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroupTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelleTestFactory;
import de.ozgcloud.eingang.router.VorgangRemoteService.VorgangCreator;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceStub;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileMetaData;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import de.ozgcloud.vorgang.vorgang.GrpcCreateVorgangRequest;
import de.ozgcloud.vorgang.vorgang.GrpcCreateVorgangResponse;
import de.ozgcloud.vorgang.vorgang.GrpcEingang;
import de.ozgcloud.vorgang.vorgang.GrpcFinishCreationRequest;
import de.ozgcloud.vorgang.vorgang.GrpcFinishCreationResponse;
import de.ozgcloud.vorgang.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;
import io.grpc.stub.CallStreamObserver;
import lombok.SneakyThrows;

class VorgangRemoteServiceTest {

	@Spy
	@InjectMocks
	private VorgangRemoteService remoteService;
	@Mock
	private VorgangManagerServerResolver resolver;
	@Mock
	private GrpcEingangMapper eingangMapper;

	private VorgangCreator vorgangCreator;

	private final FormData formData = FormDataTestFactory.create();
	private final GrpcEingang eingang = GrpcEingang.newBuilder()
			.addAttachments(GrpcIncomingFileGroupTestFactory.create())
			.addRepresentations(GrpcIncomingFileTestFactory.create())
			.build();
	private final String vorgangId = UUID.randomUUID().toString();
	private final String fileId = "42";

	@DisplayName("Create vorgang")
	@Nested
	class TestCreateVorgang {

		@Mock
		private ManagableStub<VorgangServiceBlockingStub> managableVorgangServiceStub;
		@Mock
		private VorgangServiceBlockingStub vorgangServiceStub;

		@Mock
		private ManagableStub<BinaryFileServiceStub> managableBinaryFileServiceStub;
		@Mock
		private BinaryFileServiceStub binaryFileServiceStub;

		private final Optional<String> organisationsEinheitId = Optional.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);

		@SneakyThrows
		@BeforeEach
		void mock() {
			when(resolver.resolveVorgangServiceBlockingStubByOrganisationseinheitenId(any())).thenReturn(managableVorgangServiceStub);
			when(managableVorgangServiceStub.get()).thenReturn(vorgangServiceStub);

			when(resolver.resolveBinaryFileServiceStubByOrganisationsEinheitId(any())).thenReturn(managableBinaryFileServiceStub);
			when(managableBinaryFileServiceStub.get()).thenReturn(binaryFileServiceStub);

			when(eingangMapper.toEingang(any(), any())).thenReturn(eingang);

			doNothing().when(remoteService).logConnection(any(), any());
			doNothing().when(remoteService).finishStubConnections(any());
		}

		@DisplayName("with no exception occuring")
		@Nested
		class TestWithNoException {

			@SneakyThrows
			@BeforeEach
			void mock() {
				doReturn(vorgangId).when(remoteService).createVorgang(any(), any(), any(), any());
			}

			@Test
			void shouldGetVorgangService() {
				createVorgang();

				verify(resolver).resolveVorgangServiceBlockingStubByOrganisationseinheitenId(organisationsEinheitId);
			}

			@Test
			void shouldGetBinaryFileService() {
				createVorgang();

				verify(resolver).resolveBinaryFileServiceStubByOrganisationsEinheitId(organisationsEinheitId);
			}

			@SneakyThrows
			@Test
			void shouldCreateVorgang() {
				createVorgang();

				verify(remoteService).createVorgang(formData, eingang, vorgangServiceStub, binaryFileServiceStub);
			}

			@Test
			void shouldCallEingangMapper() {
				createVorgang();

				verify(eingangMapper).toEingang(formData, organisationsEinheitId);
			}

			@Test
			void shouldFinishStubConnection() {
				createVorgang();

				verify(remoteService).finishStubConnections(List.of(managableVorgangServiceStub, managableBinaryFileServiceStub));
			}

			@Test
			void shouldReturnVorgangId() {
				var created = createVorgang();

				assertThat(created).isEqualTo(vorgangId);

			}
		}

		@DisplayName("on exception")
		@Nested
		class TestOnException {

			@SneakyThrows
			@BeforeEach
			void mock() {
				doThrow(RuntimeException.class).when(remoteService).createVorgang(any(), any(), any(), any());
			}

			@SneakyThrows
			@Test
			void shouldFinishStubConnections() {
				try {
					createVorgang();
				} catch (Exception e) {
					// Do nothing
				}

				verify(remoteService).finishStubConnections(List.of(managableVorgangServiceStub, managableBinaryFileServiceStub));
			}
		}

		@SneakyThrows
		private String createVorgang() {
			return remoteService.createVorgang(formData, organisationsEinheitId);
		}
	}

	@DisplayName("Finish stub connections")
	@Nested
	class TestFinishStubConnections {

		@Mock
		private ClosableStub<VorgangServiceBlockingStub> closableStub;

		@BeforeEach
		void mock() {
			when(closableStub.isShutdownable()).thenReturn(true);
			doNothing().when(closableStub).close();
		}

		@Test
		void shouldCheckIfSubIsShutdownable() {
			remoteService.finishStubConnections(List.of(closableStub));

			verify(closableStub).isShutdownable();
		}

		@Test
		void shouldShutDownChannelForClosableStubs() {
			remoteService.finishStubConnections(List.of(closableStub));

			verify((ClosableStub) closableStub).close();
		}
	}

	@DisplayName("VorgangCreator")
	@Nested
	class TestVorgangCreator {

		@Mock
		private final VorgangServiceBlockingStub vorgangStub = VorgangManagerServerResolverTestFactory.createVorgangBlockingStub();
		@Mock
		private final BinaryFileServiceStub binaryFileStub = VorgangManagerServerResolverTestFactory.createBinaryFileStub();
		@Mock
		private GrpcEingangMapper eingangMapper;

		@BeforeEach
		void init() {
			vorgangCreator = spy(remoteService.new VorgangCreator(formData, eingang, vorgangStub, binaryFileStub));
		}

		@Nested
		class TestCreateVorgang {

			@Mock
			private CallStreamObserver<GrpcUploadBinaryFileRequest> requestObserver;
			@Captor
			private ArgumentCaptor<GrpcCreateVorgangRequest> requestCaptor;
			@Captor
			private ArgumentCaptor<GrpcFinishCreationRequest> finishRequestCaptor;

			private final GrpcCreateVorgangResponse createVorgangResponse = GrpcCreateVorgangResponse.newBuilder().setVorgangId(vorgangId).build();
			private final GrpcFinishCreationResponse finishResponse = GrpcFinishCreationResponse.newBuilder().setMessage("OK").build();

			@BeforeEach
			void init() {
				when(vorgangStub.startCreation(any())).thenReturn(createVorgangResponse);
				when(vorgangStub.finishCreation(any())).thenReturn(finishResponse);
				doReturn(fileId).when(vorgangCreator).uploadIncomingFile(any());
				doReturn(GrpcFinishCreationRequest.newBuilder().build()).when(vorgangCreator).buildFinishCreationRequest();
			}

			@Test
			void shouldStartCreation() {
				createVorgang();

				verify(vorgangStub).startCreation(any(GrpcCreateVorgangRequest.class));
			}

			@Test
			void shouldStartCreationWithEmptyAttachmetns() {
				createVorgang();

				verify(vorgangStub).startCreation(requestCaptor.capture());
				assertThat(requestCaptor.getValue().getEingang().getAttachmentsList()).isEmpty();
			}

			@Test
			void shouldStartCreationWithEmptyRepresentations() {
				createVorgang();

				verify(vorgangStub).startCreation(requestCaptor.capture());
				assertThat(requestCaptor.getValue().getEingang().getRepresentationsList()).isEmpty();
			}

			@Test
			void shouldCallUploadAttachments() {
				createVorgang();

				verify(vorgangCreator).uploadAttachments();
			}

			@Test
			void shouldCallUploadRepresentations() {
				createVorgang();

				verify(vorgangCreator).uploadRepresentations();
			}

			@Test
			void shouldFinishCreation() {
				createVorgang();

				verify(vorgangStub).finishCreation(finishRequestCaptor.capture());
				assertThat(finishRequestCaptor.getValue()).isInstanceOf(GrpcFinishCreationRequest.class);
			}

			@Test
			void shouldReturnVorgangId() {
				var result = createVorgang();

				assertThat(result).isEqualTo(vorgangId);
			}

			@SneakyThrows
			private String createVorgang() {
				return vorgangCreator.create();
			}
		}

		@Nested
		class TestUploadAttachments {

			@BeforeEach
			void mockFileId() {
				doReturn(fileId).when(vorgangCreator).uploadIncomingFile(any());
			}

			@Test
			void shouldCallUploadIncomingFile() {
				vorgangCreator.uploadAttachments();

				verify(vorgangCreator, times(2)).uploadIncomingFile(any(IncomingFile.class));
			}

			@Test
			void shouldSetFileId() {
				var uploadedAttachments = vorgangCreator.uploadAttachments();

				assertThat(uploadedAttachments.get(0).getFiles().get(0).getId()).isEqualTo(fileId);
			}
		}

		@Nested
		class TestUploadRepresentations {

			@BeforeEach
			void mockFileId() {
				doReturn(fileId).when(vorgangCreator).uploadIncomingFile(any());
			}

			@Test
			void shouldCallUploadIncomingFile() {
				vorgangCreator.uploadRepresentations();

				verify(vorgangCreator).uploadIncomingFile(any(IncomingFile.class));
			}

			@Test
			void shouldSetFileId() {
				var uploadedRepresentations = vorgangCreator.uploadRepresentations();

				assertThat(uploadedRepresentations.get(0).getId()).isEqualTo(fileId);
			}
		}

		@Nested
		class TestBuildMetaDataRequest {

			@BeforeEach
			void mockMapper() {
				doReturn(vorgangId).when(vorgangCreator).getVorgangId();
			}

			@Test
			void shouldContainsContext() {
				var metaData = buildMetaData();

				assertThat(metaData.getContext().getClient()).isEqualTo(VorgangRemoteService.VorgangCreator.CALL_CONTEXT_CLIENT);
			}

			@Test
			void shouldContainsVorgangId() {
				var metaData = buildMetaData();

				assertThat(metaData.getVorgangId()).isEqualTo(vorgangId);
			}

			@Test
			void shouldContainsField() {
				var metaData = buildMetaData();

				assertThat(metaData.getField()).isEqualTo(VorgangRemoteService.VorgangCreator.VORGANG_ATTACHMENT_FIELD);
			}

			@Test
			void shouldContainsContentType() {
				var metaData = buildMetaData();

				assertThat(metaData.getContentType()).isEqualTo(IncomingFileTestFactory.CONTENT_TYPE);
			}

			@Test
			void shouldContainsSize() {
				var metaData = buildMetaData();

				assertThat(metaData.getSize()).isEqualTo(IncomingFileTestFactory.SIZE);
			}

			@Test
			void shouldContainsFileName() {
				var metaData = buildMetaData();

				assertThat(metaData.getFileName()).isEqualTo(IncomingFileTestFactory.NAME);
			}

			private GrpcUploadBinaryFileMetaData buildMetaData() {
				return vorgangCreator.buildMetaDataRequest(IncomingFileTestFactory.create()).getMetadata();
			}
		}

		@DisplayName("update incoming file")
		@Nested
		class TestUpdateIncomingFile {

			@Mock
			private IncomingFile incomingFile;

			@Mock
			private InputStream inputStream;

			@Mock
			private GrpcUploadBinaryFileResponse response;

			@Mock
			private GrpcUploadBinaryFileRequest request;

			@BeforeEach
			void mock() {
				doReturn(response).when(vorgangCreator).waitUntilFutureToComplete(any(), any());
				when(incomingFile.getContentStream()).thenReturn(inputStream);
				doReturn(request).when(vorgangCreator).buildMetaDataRequest(any());
			}

			@DisplayName("should call get content stream")
			@Test
			void shouldCallGetContentStream() {
				vorgangCreator.uploadIncomingFile(incomingFile);

				verify(incomingFile).getContentStream();
			}

			@DisplayName("should call build request with incoming file")
			@Test
			void shouldCallBuildRequestWithIncomingFile() {
				vorgangCreator.uploadIncomingFile(incomingFile);

				verify(vorgangCreator).buildMetaDataRequest(incomingFile);
			}

			@DisplayName("should call wait until future complete")
			@Test
			void shouldCallWaitUntilFutureComplete() {
				vorgangCreator.uploadIncomingFile(incomingFile);

				verify(vorgangCreator).waitUntilFutureToComplete(any(), eq(inputStream));
			}

		}

		@Nested
		class TestWaitUntilFutureToComplete {

			@Mock
			private FileSender<GrpcUploadBinaryFileRequest, GrpcUploadBinaryFileResponse> sender;

			@Mock
			private CompletableFuture<GrpcUploadBinaryFileResponse> streamFuture;

			@Mock
			private InputStream inputStream;

			@BeforeEach
			void initSender() {
				when(sender.getResultFuture()).thenReturn(streamFuture);
			}

			@Test
			void shouldNotThrowException() {
				assertDoesNotThrow(() -> vorgangCreator.waitUntilFutureToComplete(sender, inputStream));
			}

			@ParameterizedTest
			@ValueSource(classes = { InterruptedException.class, ExecutionException.class, TimeoutException.class })
			void shouldRethrowAsTechnicalException(Class<Exception> exception)
					throws InterruptedException, ExecutionException, TimeoutException {
				doThrow(exception).when(streamFuture).get(anyLong(), any(TimeUnit.class));

				assertThrows(TechnicalException.class, () -> vorgangCreator.waitUntilFutureToComplete(sender, inputStream));
			}

			@ParameterizedTest
			@ValueSource(classes = { InterruptedException.class, ExecutionException.class, TimeoutException.class })
			@SneakyThrows
			void shouldCloseFileContentStreamOnException(Class<Exception> exception) {
				doThrow(exception).when(streamFuture).get(anyLong(), any(TimeUnit.class));

				waitUntilFutureToComplete();

				verify(inputStream).close();
			}

			@Test
			@SneakyThrows
			void shouldCloseFileContent() {
				waitUntilFutureToComplete();

				verify(inputStream).close();
			}

			private void waitUntilFutureToComplete() {
				try {
					vorgangCreator.waitUntilFutureToComplete(sender, inputStream);
				} catch (Exception e) {
					// ignored
				}
			}
		}

		@Nested
		class TestBuildFinishCreationRequest {

			private final IncomingFileGroup attachment = IncomingFileGroupTestFactory.create();
			private final IncomingFile representation = IncomingFileTestFactory.create();

			@BeforeEach
			void mock() {
				doReturn(vorgangId).when(vorgangCreator).getVorgangId();
				doReturn(List.of(attachment)).when(vorgangCreator).getUploadedAttachments();
				doReturn(List.of(representation)).when(vorgangCreator).getUploadedRepresentations();
			}

			@Test
			void shouldContainsVorgangId() {
				var request = buildFinishCreationRequest();

				assertThat(request.getVorgangId()).isEqualTo(vorgangId);
			}

			@Test
			void shouldContainsAttachmentWithoutContent() {
				var request = buildFinishCreationRequest();

				assertThat(request.getAttachments(0).getFiles(0).getContent()).isEmpty();
			}

			@Test
			void shouldContainsRepresentationsWithoutContent() {
				var request = buildFinishCreationRequest();

				assertThat(request.getRepresentations(0).getContent()).isEmpty();
			}

			private GrpcFinishCreationRequest buildFinishCreationRequest() {
				return vorgangCreator.buildFinishCreationRequest();
			}
		}
	}
}